from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from gensim.parsing.preprocessing import (
    remove_stopwords,
    stem,
    preprocess_string,
    strip_multiple_whitespaces,
    strip_punctuation,
)
import logging
from loguru import logger
import sys
import multiprocessing

sys.path.append("./")

from refy.settings import base_dir, d2v_model_path
from refy.database import load_abstracts
from refy.progress import progress


class D2V:
    def __init__(self):
        """
            Class to load and use Doc2Vec model at inference
        """
        self.model = load_model()

    def _infer(self, input_abstract):
        """ 
            Inference part of the prediction of best matches.
            Convertes an input abstract string into its vector representation

            Arguments:
                input_abstract: str. Input abstract

            Returns:
                inferred_vector: np.ndarray
        """
        # clean up input
        words = _preprocess_string(input_abstract)

        # convert input to TaggedDocument
        input_abstract = TaggedDocument(words=words, tags=[-1])

        # infer
        return self.model.infer_vector(input_abstract.words, epochs=100)

    def predict(self, input_abstract, N=20):
        """
            Predict the best mach from the input abstract
            from the database abstracts according to the d2v model.

            Arguments:
                input_abstract: str. Input abstract
                N: int. Number of best matches to keep

            Returns:
                matches_id: list. List of indices for the best matches.
                    The indices correspond to the 
        """
        inferred_vector = self._infer(input_abstract)

        # get N best matches
        matches = self.model.docvecs.most_similar([inferred_vector], topn=N)
        matches_id = [m[0] for m in matches]

        return matches_id


# ----------------------------------- utils ---------------------------------- #
def load_model():
    """
        Loads a previously saved model
    """
    logger.debug("Loading pre-trained doc2vec model")

    return Doc2Vec.load(str(d2v_model_path))


# --------------------------------- training --------------------------------- #
def _preprocess_string(string):
    """
        Cleans up (e.g. strips punctuation) and tokenizes an input string
        to preprare for d2v processing
    """
    string = string.lower()
    filters = (
        remove_stopwords,
        stem,
        strip_multiple_whitespaces,
        strip_punctuation,
    )
    return preprocess_string(string, filters)


class Corpus:
    def __init__(self):
        """
            Class to pre-process and iterate over training data for d2v
        """
        self.training_data = load_abstracts()

    def __iter__(self):
        for ID, doc in self.training_data.items():
            yield TaggedDocument(words=_preprocess_string(doc), tags=[ID])


def train_doc2vec_model(n_epochs=50, vec_size=500, alpha=0.025):
    """
        Trains a doc2vec model from gensim for embedding and similarity 
        evaluation of paper abstracts.

        See: https://radimrehurek.com/gensim/models/doc2vec.html

        Arguments:
            n_epochs: int. Numberof epochs for training
            vec_size: int. Dimensionality of the feature vectors
            alpha: float. The initial learning rate
    """
    logger.debug("Training doc2vec model")
    logging.basicConfig(
        format="%(asctime)s : %(levelname)s : %(message)s", level=logging.INFO
    )

    # get training data
    training_data = Corpus()

    # create model
    logger.debug("Generating vocab")
    model = Doc2Vec(
        vec_size=vec_size,
        alpha=alpha,
        min_alpha=0.00025,
        min_count=2,
        sample=0,
        dm=0,
        hs=0,
        negative=5,
        workers=multiprocessing.cpu_count(),
    )
    model.build_vocab(
        training_data, progress_per=25000,
    )

    # train
    logger.debug("Training")
    with progress:
        model.train(
            training_data, total_examples=model.corpus_count, epochs=n_epochs,
        )

    # save trained
    model.save(str(base_dir / "d2v_model.model"))
    logger.debug(f"Model Saved at: {base_dir/'d2v_model.model'}")


if __name__ == "__main__":
    train_doc2vec_model()



import sys

sys.path.append("./")

from refy.database import load_database, load_abstracts
from refy import base_dir
from refy.utils import to_json


def make():
    """
        Save a small portion of the databases for faster testing
    """
    N = 10000

    # load
    db = load_database()
    abstracts = load_abstracts()

    # select
    db = db.sample(N)
    abstracts = {paper.id: abstracts[paper.id] for i, paper in db.iterrows()}

    # save
    to_json(abstracts, base_dir / "test_abstracts.json")
    db.to_hdf(base_dir / "test_database.h5", key="hdf")


if __name__ == "__main__":
    make()



from loguru import logger
import pandas as pd
import sys
import math

sys.path.append("./")

from refy.utils import (
    request,
    to_json,
)
from refy.settings import (
    fields_of_study,
    biorxiv_start_date,
    biorxiv_end_date,
    biorxiv_abstracts_path,
    biorxiv_database_path,
)
from refy.progress import progress
from refy.utils import raise_on_no_connection

"""
    Code to download papers metadata from biorxiv's API:
        https://api.biorxiv.org/
    and integrate it into refy's database
"""


# ---------------------------- make biorxiv dbase ---------------------------- #

base_url = "https://api.biorxiv.org/details/biorxiv/"


@raise_on_no_connection
def make_biorxiv_database():
    """
        Download and clean up papers metadata from biorxiv and save them to file
    """
    logger.debug(
        f'Downloading papers metadata from biorxiv: "{biorxiv_start_date}" -> "{biorxiv_end_date}"'
    )

    # first, get an idea of how many papers there are
    req = request(base_url + f"{biorxiv_start_date}/{biorxiv_end_date}")
    tot = req["messages"][0]["total"]
    logger.debug(f"Downloading metadata for {tot} papers")

    # then iterate to get 100 papers metadata at a time
    metadata = []
    abstracts = {}
    cursor = 0
    with progress:
        tsk = progress.add_task(
            "Downloading metadata from biorxiv.org", total=tot
        )

        while (
            cursor < int(math.ceil(tot / 100.0)) * 100
        ):  # round up to ensure all papers are downloaded
            # download
            data = request(
                base_url + f"{biorxiv_start_date}/{biorxiv_end_date}/{cursor}"
            )["collection"]
            cursor += 100

            # keep only relevant papers/info
            papers = pd.DataFrame(data)
            if papers.empty:
                break

            papers = papers[
                ["doi", "title", "authors", "date", "category", "abstract"]
            ]
            papers = papers.loc[papers.category.isin(fields_of_study)]
            papers["id"] = papers["doi"]
            papers = papers.drop_duplicates(subset="id")

            papers["year"] = [
                p.date.split("-")[0] for i, p in papers.iterrows()
            ]
            del papers["date"]

            # separate abstracts
            abstracts.update(
                {paper.id: paper.abstract for i, paper in papers.iterrows()}
            )
            del papers["abstract"]

            metadata.append(papers)

            # update progress
            progress.update(tsk, completed=cursor)

        # collate data and save
        papers = pd.concat(metadata)
        papers = papers.loc[papers["id"].isin(abstracts.keys())]
        papers = papers.drop_duplicates(subset="id")

        logger.debug(
            f"Found metadata for {len(papers)} relevant papers [{len(abstracts)} abstracts]"
        )

        to_json(abstracts, biorxiv_abstracts_path)
        papers.to_hdf(biorxiv_database_path, key="hdf")


if __name__ == "__main__":
    make_biorxiv_database()



from bibtexparser.bparser import BibTexParser
from pathlib import Path
import pandas as pd
from loguru import logger


def load_from_bib(fpath):
    """
        Reads from a .bib file and returns a dictionary
        with entries
    """
    parser = BibTexParser(common_strings=True)

    with open(fpath, encoding="utf-8") as bibtex_file:
        bib_database = parser.parse_file(bibtex_file)

    return bib_database.entries_dict


def load_user_input(fpath):
    """
        Parse an input library to extract authors and topics.
        From the path to a bib file extract a dictionary of bib-like entries
        and create a dataframe from these

        Arguments:
            fpath: str, Path. Path to a .bib file
    """
    # load from file
    fpath = Path(fpath)
    if fpath.suffix == ".bib":
        data = load_from_bib(fpath)
    else:
        raise NotImplementedError(
            f"Cannot parse input with file type: {fpath.suffix}"
        )
    logger.debug(f"Loaded user input from file: {fpath} | {len(data)} entries")

    # Clean up data
    data = pd.DataFrame(data.values())
    data = data[["title", "journal", "author", "abstract", "doi"]]
    data.columns = ["title", "journal", "authors", "abstract", "doi"]
    data["id"] = data["title"]

    # keep only papers with abstract
    has_abs = [
        True if (isinstance(a, str) and len(a) > 1) else False
        for a in data["abstract"].values
    ]

    data = data[has_abs].copy()
    data["input"] = True

    # log and return
    logger.debug(f"{len(data)}/{len(has_abs)} user papers have abstracts")

    return data



import json
import pandas as pd
from pathlib import Path
import gzip
import multiprocessing
from loguru import logger
from rich.progress import track
from langdetect import detect
from rich import print

import sys

sys.path.append("./")

from refy.settings import (
    fields_of_study,
    low_year,
    keywords,
    database_path,
    abstracts_path,
)
from refy.utils import (
    isin,
    to_json,
    from_json,
)

ABSTRACTS = {}  # store all abstracts before saving

# ---------------------------- preprocess database --------------------------- #


def exclude(entry):
    """
        Select papers based on:
            * field of study
            * publication data
            * if they include keywords in their abstract
            * they are written in english
            * if they have an abstract

        the parameters are set in settings.py

        Arguments:
            entry: dict with paper's metadata

        Returns:
            exclude: bool. True if the entry fails any of the criteria
    """
    # keep only entries in relevant fields
    if not isin(entry["fieldsOfStudy"], fields_of_study):
        return True

    # Keep only recent papers
    entry["year"] = entry["year"] or 0
    if entry["year"] < low_year:
        return True

    # Keep only entries that have an abstract
    abstract = entry["paperAbstract"].lower()
    if len(abstract) < 1:
        return True

    # keep only entries with keywords in abstract
    if not any((keyword in abstract) for keyword in keywords):
        return True

    # keep only english
    try:
        lang = detect(abstract[:50])
    except Exception:
        return True
    else:
        if lang != "en":
            return True

    # ok all good
    return False


def _unpack_single_file(args):
    """
        Unpacks a single .gz file and filters it
        to only keep relevant papers.
        The (ugly) way the arguments are passed to this
        function is to facilitate multiprocessing
    """
    fpath, dfs_dir, n, N = args

    # prepare paths
    name = fpath.name.split(".")[0]
    out = dfs_dir / (name + ".h5")
    out_abs = dfs_dir / (name + "_abstracts.json")

    # load data
    if fpath.suffix == ".gz":
        with gzip.open(fpath, "r") as datafile:
            data = datafile.readlines()
        data = [d.decode("utf-8") for d in data]
    else:
        with open(fpath) as datafile:
            data = datafile.readlines()

    # create a dataframe with relevant data
    metadata = dict(title=[], authors=[], doi=[], url=[], id=[], year=[])

    # loop over all entries
    abstracts = {}
    for entry in data:
        entry = json.loads(entry)

        if exclude(entry):
            continue

        # store abstract
        abstracts[entry["id"]] = entry["paperAbstract"]

        # keep metadata
        metadata["year"].append(
            str(entry["year"]) if "year" in entry.keys() else ""
        )
        metadata["id"].append(str(entry["id"]))
        metadata["title"].append(str(entry["title"]))
        metadata["authors"].append(
            ", ".join([str(a["name"]) for a in entry["authors"]])
        )
        metadata["doi"].append(entry["doi"] or "")
        metadata["url"].append(entry["s2Url"] or "")

    # save
    metadata = pd.DataFrame(metadata)
    metadata.to_hdf(out, key="hdf")

    to_json(abstracts, out_abs)

    logger.debug(
        f"Uncompressed: {fpath.name} [{n}/{N}]. Kept {len(metadata)}/{len(data)} papers and {len(abstracts)} abstracts"
    )


def upack_database(folder):
    """
        Opens up .gz with papers metadata info from 
        http://s2-public-api-prod.us-west-2.elasticbeanstalk.com/corpus/download/
        and saves is at a .h5 file. 
        The operation is parallelized to speed things up. 

        For each .gz file:
            1. open the file and load the data
            2. select papers that match the criteria set in settings.py
            3. save the selected papers' metadata to .h5 (pandas dataframe) in folder/dfs
            4. save the selcted papers's abstracct to .txt in folder/abstracts

        Arguments:
            folder: str, Path. Path to the folder where the database data will be stored.
                It must include a subfolder called 'compressed' in which the .gz files live. 
    """
    logger.debug(f"Unpacking database in {folder}")

    # get folders
    folder = Path(folder)

    dfs_dir = folder / "dfs"
    dfs_dir.mkdir(exist_ok=True)

    # extract data from all files
    files = list((folder / "compressed").glob("*.gz"))
    logger.debug(f"Uncompressing {len(files)} files")

    n_cpus = multiprocessing.cpu_count() - 2
    with multiprocessing.Pool(processes=n_cpus) as pool:
        args = [(fl, dfs_dir, n, len(files)) for n, fl in enumerate(files)]
        pool.map(_unpack_single_file, args, chunksize=1)


def make_database(folder):
    """
        Given a database folder filled in by `unpack_database` 
        this function creates the database proper. 

        Arguments:
            folder: str, Path. Path to the folder where the database data is stored.
                User must have run `unpack_database` on the folder's content first. 
    """
    logger.debug(f"Making database from data folder: {folder}")

    folder = Path(folder)
    files = list((folder / "dfs").glob("*.h5"))
    logger.debug(f"Found {len(files)} files")

    # Load all metadata into a single dataframe
    dfs = []
    for f in track(files, description="Loading data..."):
        dfs.append(pd.read_hdf(f, key="hdf"))

    # concatenate
    DATA = pd.concat(dfs)
    logger.debug(f"Found {len(DATA)} papers")
    print(
        "\n\n[b green]Titles samples:",
        *DATA.title.values[:15],
        "\n",
        sep="\n\n",
    )

    # remove duplicates
    DATA = DATA.drop_duplicates(subset="title")
    logger.debug(f"After dropping duplicates  {len(DATA)} papers are left")

    # load and save abstracts
    logger.debug("Loading abstracts")
    abs_list = [from_json(f) for f in (folder / "dfs").glob("*.json")]
    abstracts = {k: v for d in abs_list for k, v in d.items()}
    logger.debug(f"Loaded {len(abstracts)} abstracts")

    # keep only papers that have an abstract
    logger.debug("Cleaning up data")
    DATA = DATA.loc[DATA["id"].isin(abstracts.keys())]

    # save data
    logger.debug("Saving database")
    DATA.to_hdf(database_path, key="hdf")
    logger.debug(
        f"Saved database at: {database_path}. {len(DATA)} entries in total"
    )

    # and only abstracts that have a paper
    logger.debug("Cleaning up abstracts")
    cleaned_abstracts = {k: abstracts[k] for k in DATA["id"].values}

    if len(cleaned_abstracts) != len(DATA):
        raise ValueError(
            f"Found {len(cleaned_abstracts)} abstracts and {len(DATA)} papers !!"
        )

    logger.debug(
        f"Saving {len(cleaned_abstracts)} abstracts to file: {abstracts_path}"
    )
    to_json(cleaned_abstracts, abstracts_path)


if __name__ == "__main__":
    fld = "M:\\PAPERS_DBASE"
    # upack_database(fld)

    make_database(fld)



import pandas as pd
from loguru import logger

from refy.settings import (
    database_path,
    abstracts_path,
    biorxiv_abstracts_path,
    biorxiv_database_path,
    test_abstracts_path,
    test_database_path,
)
from refy import settings
from refy.utils import from_json


# --------------------------------- load data -------------------------------- #


def load_abstracts():
    """
        loads all abstracts from the json files
    
        Returns:
            abstracts: dict with all abstracts
    """
    if not settings.TEST_MODE:  # pragma: no cover
        logger.debug("Loading abstracts")

        abstracts = from_json(abstracts_path)
        abstracts.update(from_json(biorxiv_abstracts_path))
    else:
        logger.debug("loading abstracts in TEST MODE")
        abstracts = from_json(test_abstracts_path)

    # remove empy abstracts
    logger.debug(f"Loaded {len(abstracts)} abstracts")
    return abstracts


def load_database():
    """
        Load papers databases from files.
        
        Returns:
            database: DataFrame with search database metadata
    """
    if not settings.TEST_MODE:  # pragma: no cover
        logger.debug("Loading database")

        # download semanthic scholar database
        dbase = pd.read_hdf(database_path, key="hdf")
        dbase["source"] = "semanthic scholar"

        # download biorxiv database
        biorxiv_dbase = pd.read_hdf(biorxiv_database_path, key="hdf")
        del biorxiv_dbase["category"]
        biorxiv_dbase["source"] = "biorxiv"

        # merge
        dbase = pd.concat([dbase, biorxiv_dbase], sort=True).reset_index()
    else:
        logger.debug(f"Loading database in TEST MODE")
        dbase = pd.read_hdf(test_database_path, key="hdf")

    # clean up years column
    years = [int(p.year) if p.year else 0 for i, p in dbase.iterrows()]
    dbase["year"] = years

    dbase["input"] = False  # differentiate from user input

    logger.debug(f"Loaded database with {len(dbase)} entries")
    return dbase



from loguru import logger
from rich import print
import sys
from pathlib import Path
from myterial import orange

sys.path.append("./")
from refy.input import load_user_input
from refy.database import load_database
from refy.progress import suggest_progress
from refy.utils import to_table
from refy import doc2vec as d2v
from refy import download


def suggest_one(input_string, N=20, since=None, to=None, savepath=None):
    """
        Finds recomendations based on a single input string (with keywords,
        or a paper abstract or whatever) instead of an input .bib file

        Arguments:
            input_stirng: str. String to match against database
            N: int. Number of papers to suggest
            since: int or None. If an int is passed it must be a year,
                only papers more recent than the given year are kept for recomendation
            to: int or None. If an int is passed it must be a year,
                only papers older than that are kept for recomendation
            savepath: str, Path. Path pointing to a .csv file where the recomendations
                will be saved

        Returns:
            suggestions: pd.DataFrame of N recomended papers
    """
    download.check_files()

    # load database and abstracts
    database = load_database()

    # load model
    model = d2v.D2V()

    # find recomendations
    best_IDs = model.predict(input_string, N=N)

    # get selected papers
    suggestions = database.loc[database["id"].isin(best_IDs)].reset_index()
    suggestions["score"] = None

    # select papers based on date of publication
    if since:
        suggestions = suggestions.loc[suggestions.year >= int(since)]
    if to:
        suggestions = suggestions.loc[suggestions.year <= int(to)]

    # print
    print(
        to_table(
            suggestions,
            title=f'Suggestions for input string: [bold {orange}]"{input_string}"',
        )
    )

    # save to file
    if savepath:
        suggestions.to_csv(savepath)

    return suggestions


class suggest:
    suggestions_per_paper = 100  # for each paper find N suggestions

    def __init__(self, user_papers, N=20, since=None, to=None, savepath=None):
        """
            Suggest new relevant papers based on the user's
            library.

            Arguments:
                user_papers: str, path. Path to a .bib file with user's papers info
                N: int. Number of papers to suggest
                since: int or None. If an int is passed it must be a year,
                    only papers more recent than the given year are kept for recomendation
                to: int or None. If an int is passed it must be a year,
                    only papers older than that are kept for recomendation
                savepath: str, Path. Path pointing to a .csv file where the recomendations
                    will be saved
        """
        download.check_files()

        self.since = since
        self.to = to
        if savepath:
            self.savepath = Path(savepath)
        else:
            self.savepath = savepath

        with suggest_progress as progress:
            self.progress = progress
            self.n_completed = -1
            self.task_id = self.progress.add_task(
                "Suggesting papers..", start=True, total=4, current_task="",
            )
            # load data
            self.load_data(user_papers)

            # load d2v model
            self._progress("Loading Doc2Vec model")
            self.d2v = d2v.D2V()

            # get suggestions
            self.get_suggestions(N=N)

    @property
    def n_papers(self):
        return len(self.database)

    @property
    def n_user_papers(self):
        return len(self.user_papers)

    def _progress(self, task_name):
        """
            Update progress bar
        """
        self.n_completed += 1
        self.progress.update(
            self.task_id, current_task=task_name, completed=self.n_completed
        )

    def load_data(self, user_papers):
        """
            Load papers metadata for user papers and database papers

            Arguments:
                user_papers: str, path. Path to a .bib file with user's papers info
        """
        # load database
        self._progress("Loading database papers")
        self.database = load_database()

        # load user data
        self._progress("Loading user papers",)
        self.user_papers = load_user_input(user_papers)

    def suggest_for_paper(self, user_paper_title, user_paper_abstract):
        """
            Finds the best matches for a single paper

            Arguments:
                user_paper_title: str. Title of input user paper
                user_paper_abstract: str. Abstract of input user paper

            Returns:
                suggestions: dict. Dictionary of title:value where value 
                    is self.suggestions_per_paper for the best match paper and 
                    1 for the lowest match
        """
        # find best match with d2v
        best_IDs = self.d2v.predict(
            user_paper_abstract, N=self.suggestions_per_paper
        )

        # get selected papers
        selected = self.database.loc[self.database["id"].isin(best_IDs)]

        if selected.empty:
            logger.debug(
                f'Could not find any suggested papers for paper: "{user_paper_title}" '
            )

        return {
            t: self.suggestions_per_paper - n
            for n, t in enumerate(selected.title.values)
        }

    def _collate_suggestions(self, points):
        """
            Given a dictionart of points for each suggested paper,
            this function returns a dataframe with papers ordered 
            by their score

            Arguments:
                points: dict of title:point entries for each recomended paper

            Returns
                suggestions: pd.DataFrame with suggested papers ordred by score
        """
        # collate recomendations
        suggestions = self.database.loc[
            self.database.title.isin(points.keys())
        ].drop_duplicates(subset="title")

        # drop suggestions whose title is in the user papers
        L = len(suggestions)
        suggestions = suggestions.loc[
            ~suggestions.title.isin(self.user_papers.title)
        ]
        suggestions = suggestions.loc[
            ~suggestions.doi.isin(self.user_papers.doi)
        ]
        logger.debug(
            f"While collating suggestions, discarded {len(suggestions)-L} suggestions found among user inputs"
        )

        # Get each paper's score
        max_score = self.suggestions_per_paper * self.n_user_papers
        suggestions["score"] = [
            points[title] / max_score for title in suggestions.title.values
        ]

        # sort recomendations based on score
        suggestions = suggestions.sort_values("score", ascending=False)

        # keep only papers published within a given years range
        if self.since:
            suggestions = suggestions.loc[suggestions.year >= int(self.since)]
        if self.to:
            suggestions = suggestions.loc[suggestions.year <= int(self.to)]

        suggestions = suggestions.reset_index(drop=True)

        return suggestions

    def get_suggestions(self, N=20):
        """
            Finds the papers from the database that are not in the user's
            library but are most similar to the users papers.
            For each user paper, get the N most similar papers, then get
            the papers that came up most frequently across all user papers.
            

            Arguments:
                N: int, number of best papers to keep

            Returns:
                suggestions: pd.DataFrame with suggested papers sorted by score
        """
        logger.debug(f"Getting suggestions for {self.n_user_papers} papers")

        # progress bar
        self._progress("Finding matches")
        select_task = self.progress.add_task(
            "Selecting best matches...",
            start=True,
            total=self.n_user_papers,
            current_task="analyzing...",
        )

        # find best matches for each paper
        points = {}
        for n, (idx, user_paper) in enumerate(self.user_papers.iterrows()):
            # keep track of recomendations across all user papers
            paper_suggestions = self.suggest_for_paper(
                user_paper.title, user_paper.abstract
            )

            for suggested, pts in paper_suggestions.items():
                if suggested in points.keys():
                    points[suggested] += pts
                else:
                    points[suggested] = pts

            self.progress.update(select_task, completed=n)
        self.progress.remove_task(select_task)
        self.progress.remove_task(self.task_id)

        # collate and print suggestions
        self.suggestions = self._collate_suggestions(points)[:N]
        print(to_table(self.suggestions))

        # save to file
        if self.savepath:
            self.suggestions.to_csv(self.savepath)

        return self.suggestions


if __name__ == "__main__":
    import refy

    refy.set_logging("DEBUG")
    suggest(refy.settings.example_path, N=100, since=2018)

    # suggest_one("locomotion control")



from refy import suggest, suggest_one
from refy.settings import example_path, base_dir
import pandas as pd


def test_suggst_one():
    suggestions = suggest_one(
        "neuron gene expression brain nervous system", N=20
    )

    assert isinstance(suggestions, pd.DataFrame), "returned wrong class"

    # suggestions = suggest_one("neuron gene expression brain nervous system", N=20, since=2015, to=2018)

    # assert suggestions.year.min() == 2015
    # assert suggestions.year.max() == 2018


def test_suggest_save():
    # create a path to save the suggestions to
    save_path = base_dir / "ref_test.csv"
    if save_path.exists():
        save_path.unlink()

    # get suggestions
    suggestions = suggest(example_path, N=20, savepath=save_path).suggestions

    # check suggestions
    assert isinstance(suggestions, pd.DataFrame), "wrong class"
    assert len(suggestions) == 20, "wrong number of sugg"

    # check saved suggestions
    assert save_path.exists(), "didnt save"
    saved = pd.read_csv(save_path)

    assert len(saved) == 20, "loaded has wrong length"

    save_path.unlink()


def test_suggest_criteria():

    # test criteria
    suggestions = suggest(example_path, N=20, since=2015, to=2019).suggestions

    assert suggestions.year.min() == 2015, "since doesnt work"
    assert suggestions.year.max() == 2019, "to doesnt work"



from pyinspect import install_traceback

install_traceback(hide_locals=True)

from refy.suggest import suggest, suggest_one
from refy.settings import (
    DEBUG,
    base_dir,
)


# ----------------------------- logging settings ----------------------------- #

from loguru import logger
import sys


def set_logging(level="INFO"):
    logger.remove()
    logger.add(sys.stdout, level=level)
    logger.add(str(base_dir / "log.log"), level="DEBUG")


if not DEBUG:
    set_logging()
else:
    set_logging(level="DEBUG")



import json
import requests
from rich.table import Table
from rich import print
from rich.panel import Panel

from myterial import salmon, orange, amber

# ------------------------------- print papers ------------------------------- #


def to_table(papers, title=None):
    """
        prints a dataframe with papers metadata in a pretty format

        Arguments:
            papers: pd.DataFrame
            title: str. Optional input to replace default title

    """
    if papers.empty:
        print(f"[{orange}]Found no papers matching your query, sorry")
        return

    # create table
    table = Table(
        show_header=True,
        header_style="bold white",
        show_lines=True,
        expand=False,
        box=None,
        title=title or "Recomended papers",
        title_style=f"bold {salmon}",
        caption=f"{len(papers)} papers, recommended by refy :ok_hand:",
        caption_style="dim",
        padding=(0, 1),
    )
    table.add_column("#")
    table.add_column(":thumbs_up: score", style="dim", justify="left")
    table.add_column("year", style="dim", justify="center")
    table.add_column("title", style=f"bold {orange}", min_width=40)
    table.add_column(
        "DOI/url", style="dim",
    )

    # add papers to table
    for i, paper in papers.iterrows():
        if paper.score:
            score = "[dim]" + str(round(paper["score"], 3))
        else:
            score = ""

        if paper.doi:
            url = f'[link=https://doi.org/{paper.doi}]https://doi.org/{paper.doi}[/]"'
        else:
            url = f"[link={paper.url}]{paper.url}"

        table.add_row(
            str(i + 1),
            score,
            f"[dim {amber}]" + str(paper.year),
            paper.title,
            url,
        )

    # fit in a panel
    return Panel(
        table, expand=False, border_style=f"{orange}", padding=(0, 2, 1, 2)
    )


def isin(l1, l2):
    """
        Checks if any element of a list is included in a second list
    """
    return any(x in l2 for x in l1)


# --------------------------------- internet --------------------------------- #
def check_internet_connection(
    url="http://www.google.com/", timeout=2, raise_error=True
):
    """Check that there is an internet connection
    url : str
        url to use for testing (Default value = 'http://www.google.com/')
    timeout : int
        timeout to wait for [in seconds] (Default value = 2).
    raise_error : bool
        if false, warning but no error.
    """

    try:
        _ = requests.get(url, timeout=timeout)
        return True
    except requests.ConnectionError:
        if not raise_error:
            print("No internet connection available.")
        else:
            raise ConnectionError(
                "No internet connection, try again when you are connected to the internet."
            )
    return False


def raise_on_no_connection(func):
    """
        Decorator to avoid running a function when there's no internet
    """

    def inner(*args, **kwargs):
        if not check_internet_connection():
            raise ConnectionError("No internet connection found.")
        else:
            return func(*args, **kwargs)

    return inner


def _request(url, stream=False):
    """
        Sends a request to an url and
        makes sure it worked
    """
    response = requests.get(url, stream=stream)
    if not response.ok:
        raise ValueError(
            f"Failed to get a good response when retrieving from {url}. Response: {response.status_code}"
        )
    return response


@raise_on_no_connection
def request(url):
    """ 
        Sends a request to a URL and returns the JSON
        it fetched (if it went through).
    """
    response = _request(url, stream=False)
    return response.json()


# --------------------------------- File I/O --------------------------------- #


def to_json(obj, fpath):
    """ saves an object to json """
    with open(fpath, "w") as out:
        json.dump(obj, out)


def from_json(fpath):
    """ loads an object from json """
    with open(fpath, "r") as fin:
        return json.load(fin)
